#!/usr/bin/env python3
import shutil
import site
from pathlib import Path
from sys import argv

import torch
from predtuner import config_pylogger
from predtuner.pipedbin import PipedBinaryApp
from torch2hpvm import BinDataset, ModelExporter
from torch.nn import Module

site.addsitedir(Path(__file__).parent.parent.absolute())
import dnn

# Set up logger
msg_logger = config_pylogger(output_dir=".", verbose=True)
benchmarks = {
    "lenet_mnist": (dnn.LeNet, 1, 28, 1000),
    "alexnet_imagenet": (dnn.AlexNetImageNet, 3, 224, 100),
    "mobilenet_cifar10": (dnn.MobileNet, 3, 32, 500),
    "resnet18_cifar10": (dnn.ResNet18, 3, 32, 500),
    "vgg16_cifar10": (dnn.VGG16Cifar10, 3, 32, 500),
}
model_param = Path(__file__).parent / "../../model_params"


def generate(model_cls, nch, img_size, batch_size, netname):
    codegen_dir = Path(f"./{netname}")
    build_dir = codegen_dir / "build"
    metadata_file = codegen_dir / "ops.json"
    binary_file = build_dir / netname
    build_dir = codegen_dir / "build"

    if codegen_dir.exists():
        shutil.rmtree(codegen_dir)
    params = model_param / netname
    dataset_shape = 5000, nch, img_size, img_size
    bin_tuneset = BinDataset(
        params / "tune_input.bin", params / "tune_labels.bin", dataset_shape
    )
    bin_testset = BinDataset(
        params / "test_input.bin", params / "test_labels.bin", dataset_shape
    )
    model: Module = model_cls()
    checkpoint = model_param / f"pytorch/{netname}.pth.tar"
    model.load_state_dict(torch.load(checkpoint.as_posix()))
    exporter = ModelExporter(
        model, bin_tuneset, bin_testset, codegen_dir, target="hpvm_tensor_inspect"
    )
    exporter.generate(batch_size=batch_size).compile(binary_file, build_dir)
    return binary_file, metadata_file


def main():
    netname, is_pred = argv[1:]
    is_pred = int(is_pred)
    model_cls, nch, img_size, batch_size = benchmarks[netname]
    binary_file, metadata_file = generate(
        model_cls, nch, img_size, batch_size, netname
    )
    app = PipedBinaryApp("test", binary_file, metadata_file)
    tuner = app.get_tuner()
    tuner.tune(
        5,
        3.0,
        is_threshold_relative=True,
        cost_model="cost_linear",
        qos_model="qos_p1" if is_pred else "none",
    )
    tuner.dump_configs("configs.json")
    fig = tuner.plot_configs(show_qos_loss=True)
    fig.savefig("configs.png", dpi=300)
    app.dump_hpvm_configs(tuner.best_configs, "hpvm_confs.txt")


if __name__ == "__main__":
    main()
